from django.apps import AppConfig


class AnomalyTestsConfig(AppConfig):
    name = 'ambition_utils.anomaly.tests'
    verbose_name = 'Ambition Anomaly Tests'
    label = 'ambition_utils_anomaly_tests'
