# Generated by Django 2.2.28 on 2022-06-23 16:49

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('rrule', '0002_auto_20190422_2037'),
    ]

    operations = [
        migrations.AddField(
            model_name='rrule',
            name='related_object_content_type',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.PROTECT, to='contenttypes.ContentType'),
        ),
        migrations.AddField(
            model_name='rrule',
            name='related_object_handler_name',
            field=models.TextField(blank=True, default=None, null=True),
        ),
        migrations.AddField(
            model_name='rrule',
            name='related_object_id',
            field=models.IntegerField(db_index=True, null=True),
        ),
    ]
