# Generated by Django 3.2.20 on 2023-08-02 15:48

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('rrule', '0004_rrule_rrule_exclusion_params'),
    ]

    operations = [
        migrations.AddField(
            model_name='rrule',
            name='day_offset',
            field=models.SmallIntegerField(blank=True, null=True),
        ),
    ]
