# Generated by Django 2.2.28 on 2023-01-06 21:13

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('rrule', '0003_auto_20220623_1649'),
    ]

    operations = [
        migrations.CreateModel(
            name='Program',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.TextField()),
                ('start_called', models.IntegerField(default=0)),
                ('end_called', models.IntegerField(default=0)),
                ('end_recurrence', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='programs_end', to='rrule.RRule')),
                ('start_recurrence', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='programs_start', to='rrule.RRule')),
            ],
        ),
    ]
