# Generated by Django 2.2.28 on 2023-01-06 21:15

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='FloatModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('value', models.FloatField()),
            ],
        ),
        migrations.CreateModel(
            name='IncrementalAnomaly',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('uid', models.CharField(max_length=256, unique=True)),
                ('blob', models.JSONField(default=dict, null=True)),
                ('percentile_low', models.FloatField(null=True)),
                ('percentile_high', models.FloatField(null=True)),
                ('delta', models.FloatField(default=0.01, null=True)),
                ('K', models.FloatField(default=25, null=True)),
                ('threshold_low', models.FloatField(db_index=True, null=True)),
                ('threshold_high', models.FloatField(db_index=True, null=True)),
                ('num_values_ingested', models.IntegerField(default=0)),
                ('last_modified', models.DateTimeField(db_index=True)),
                ('name', models.CharField(max_length=100)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='NonIncrementalAnomaly',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('uid', models.CharField(max_length=256, unique=True)),
                ('blob', models.JSONField(default=dict, null=True)),
                ('percentile_low', models.FloatField(null=True)),
                ('percentile_high', models.FloatField(null=True)),
                ('delta', models.FloatField(default=0.01, null=True)),
                ('K', models.FloatField(default=25, null=True)),
                ('threshold_low', models.FloatField(db_index=True, null=True)),
                ('threshold_high', models.FloatField(db_index=True, null=True)),
                ('num_values_ingested', models.IntegerField(default=0)),
                ('last_modified', models.DateTimeField(db_index=True)),
                ('name', models.CharField(max_length=100)),
            ],
            options={
                'abstract': False,
            },
        ),
    ]
