# Generated by Django 3.2.13 on 2022-05-25 17:54

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('postgres_lock', '0002_auto_20191030_1417'),
    ]

    operations = [
        migrations.AddField(
            model_name='postgreslock',
            name='previous_value',
            field=models.TextField(default=None, null=True),
        ),
        migrations.AddField(
            model_name='postgreslock',
            name='value',
            field=models.TextField(default=None, null=True),
        ),
    ]
