# Generated by Django 3.2.19 on 2023-06-05 19:55

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('rrule', '0003_auto_20220623_1649'),
    ]

    operations = [
        migrations.AddField(
            model_name='rrule',
            name='rrule_exclusion_params',
            field=models.JSONField(blank=True, default=None, null=True),
        ),
    ]
