# Generated by Django 2.0.6 on 2018-06-04 19:00

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='PostgresLock',
            fields=[
                ('name', models.CharField(max_length=255, primary_key=True, serialize=False, unique=True)),
                ('time', models.DateTimeField(default=None)),
            ],
        ),
    ]
