# Generated by Django 2.2.4 on 2019-10-30 14:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('postgres_lock', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='postgreslock',
            name='name',
            field=models.CharField(max_length=512, primary_key=True, serialize=False, unique=True),
        ),
    ]
