from django.apps import AppConfig


class RRuleTestsConfig(AppConfig):
    name = 'ambition_utils.rrule.tests'
    verbose_name = 'Ambition Utils RRule Tests'
    label = 'ambition_utils_rrule_tests'
